@@ -25,14 +25,16 @@
 
 #if PY_MAJOR_VERSION >= 3
 
+#define MyText_Type         PyUnicode_Type
 #define MyText_Check(o)     PyUnicode_Check(o)
-#define MyText_AS_STRING(o) FOOEY_DONT_KNOW_YET(o)
+#define MyText_AS_STRING(o) PyBytes_AS_STRING(PyUnicode_AsASCIIString(o))
 #define MyInt_FromLong(l)   PyLong_FromLong(l)
 
 #define MyType_HEAD_INIT    PyVarObject_HEAD_INIT(NULL, 0)
 
 #else
 
+#define MyText_Type         PyString_Type
 #define MyText_Check(o)     PyString_Check(o)
 #define MyText_AS_STRING(o) PyString_AS_STRING(o)
 #define MyInt_FromLong(l)   PyInt_FromLong(l)
@@ -41,6 +43,10 @@
 
 #endif /* Py3k */
 
+/* The values returned to indicate ok or error. */
+#define RET_OK      0
+#define RET_ERROR   -1
+
 /* An entry on the data stack.  For each call frame, we need to record the
     dictionary to capture data, and the last line number executed in that
     frame.
@@ -93,6 +99,7 @@ typedef struct {
 
     /* The parent frame for the last exception event, to fix missing returns. */
     PyFrameObject * last_exc_back;
+    int last_exc_firstlineno;
 
 #if COLLECT_STATS
     struct {
@@ -112,7 +119,7 @@ typedef struct {
 #define STACK_DELTA    100
 
 static int
-Tracer_init(Tracer *self, PyObject *args, PyObject *kwds)
+Tracer_init(Tracer *self, PyObject *args_unused, PyObject *kwds_unused)
 {
 #if COLLECT_STATS
     self->stats.calls = 0;
@@ -139,7 +146,7 @@ Tracer_init(Tracer *self, PyObject *args, PyObject *kwds)
     if (self->data_stack == NULL) {
         STATS( self->stats.errors++; )
         PyErr_NoMemory();
-        return -1;
+        return RET_ERROR;
     }
     self->data_stack_alloc = STACK_DELTA;
 
@@ -148,7 +155,7 @@ Tracer_init(Tracer *self, PyObject *args, PyObject *kwds)
 
     self->last_exc_back = NULL;
 
-    return 0;
+    return RET_OK;
 }
 
 static void
@@ -219,7 +226,7 @@ static const char * what_sym[] = {"CALL", "EXC ", "LINE", "RET "};
 static int
 Tracer_record_pair(Tracer *self, int l1, int l2)
 {
-    int ret = 0;
+    int ret = RET_OK;
 
     PyObject * t = PyTuple_New(2);
     if (t != NULL) {
@@ -227,13 +234,13 @@ Tracer_record_pair(Tracer *self, int l1, int l2)
         PyTuple_SET_ITEM(t, 1, MyInt_FromLong(l2));
         if (PyDict_SetItem(self->cur_file_data, t, Py_None) < 0) {
             STATS( self->stats.errors++; )
-            ret = -1;
+            ret = RET_ERROR;
         }
         Py_DECREF(t);
     }
     else {
         STATS( self->stats.errors++; )
-        ret = -1;
+        ret = RET_ERROR;
     }
     return ret;
 }
@@ -242,9 +249,9 @@ Tracer_record_pair(Tracer *self, int l1, int l2)
  * The Trace Function
  */
 static int
-Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
+Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg_unused)
 {
-    int ret = 0;
+    int ret = RET_OK;
     PyObject * filename = NULL;
     PyObject * tracename = NULL;
 
@@ -276,8 +283,8 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
             STATS( self->stats.missed_returns++; )
             if (self->depth >= 0) {
                 if (self->tracing_arcs && self->cur_file_data) {
-                    if (Tracer_record_pair(self, self->last_line, -1) < 0) {
-                        return -1;
+                    if (Tracer_record_pair(self, self->last_line, -self->last_exc_firstlineno) < 0) {
+                        return RET_ERROR;
                     }
                 }
                 SHOWLOG(self->depth, frame->f_lineno, frame->f_code->co_filename, "missedreturn");
@@ -304,7 +311,7 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
                 STATS( self->stats.errors++; )
                 PyErr_NoMemory();
                 self->depth--;
-                return -1;
+                return RET_ERROR;
             }
             self->data_stack = bigger_data_stack;
             self->data_stack_alloc = bigger;
@@ -327,11 +334,11 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
             if (tracename == NULL) {
                 /* An error occurred inside should_trace. */
                 STATS( self->stats.errors++; )
-                return -1;
+                return RET_ERROR;
             }
             if (PyDict_SetItem(self->should_trace_cache, filename, tracename) < 0) {
                 STATS( self->stats.errors++; )
-                return -1;
+                return RET_ERROR;
             }
         }
         else {
@@ -345,13 +352,13 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
                 file_data = PyDict_New();
                 if (file_data == NULL) {
                     STATS( self->stats.errors++; )
-                    return -1;
+                    return RET_ERROR;
                 }
                 ret = PyDict_SetItem(self->data, tracename, file_data);
                 Py_DECREF(file_data);
                 if (ret < 0) {
                     STATS( self->stats.errors++; )
-                    return -1;
+                    return RET_ERROR;
                 }
             }
             self->cur_file_data = file_data;
@@ -372,8 +379,9 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
         /* A near-copy of this code is above in the missing-return handler. */
         if (self->depth >= 0) {
             if (self->tracing_arcs && self->cur_file_data) {
-                if (Tracer_record_pair(self, self->last_line, -1) < 0) {
-                    return -1;
+                int first = frame->f_code->co_firstlineno;
+                if (Tracer_record_pair(self, self->last_line, -first) < 0) {
+                    return RET_ERROR;
                 }
             }
 
@@ -393,7 +401,7 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
                 if (self->tracing_arcs) {
                     /* Tracing arcs: key is (last_line,this_line). */
                     if (Tracer_record_pair(self, self->last_line, frame->f_lineno) < 0) {
-                        return -1;
+                        return RET_ERROR;
                     }
                 }
                 else {
@@ -401,13 +409,13 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
                     PyObject * this_line = MyInt_FromLong(frame->f_lineno);
                     if (this_line == NULL) {
                         STATS( self->stats.errors++; )
-                        return -1;
+                        return RET_ERROR;
                     }
                     ret = PyDict_SetItem(self->cur_file_data, this_line, Py_None);
                     Py_DECREF(this_line);
                     if (ret < 0) {
                         STATS( self->stats.errors++; )
-                        return -1;
+                        return RET_ERROR;
                     }
                 }
             }
@@ -432,6 +440,7 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
         */
         STATS( self->stats.exceptions++; )
         self->last_exc_back = frame->f_back;
+        self->last_exc_firstlineno = frame->f_code->co_firstlineno;
         break;
 
     default:
@@ -439,22 +448,61 @@ Tracer_trace(Tracer *self, PyFrameObject *frame, int what, PyObject *arg)
         break;
     }
 
-    return 0;
+    return RET_OK;
+}
+
+/*
+ * A sys.settrace-compatible function that invokes our C trace function.
+ */
+static PyObject *
+Tracer_pytrace(Tracer *self, PyObject *args)
+{
+    PyFrameObject *frame;
+    PyObject *what_str;
+    PyObject *arg_unused;
+    int what;
+    static char *what_names[] = {
+        "call", "exception", "line", "return",
+        "c_call", "c_exception", "c_return",
+        NULL
+        };
+
+    if (!PyArg_ParseTuple(args, "O!O!O:Tracer_pytrace",
+            &PyFrame_Type, &frame, &MyText_Type, &what_str, &arg_unused)) {
+        goto done;
+    }
+
+    /* In Python, the what argument is a string, we need to find an int
+       for the C function. */
+    for (what = 0; what_names[what]; what++) {
+        if (!strcmp(MyText_AS_STRING(what_str), what_names[what])) {
+            break;
+        }
+    }
+
+    /* Invoke the C function, and return ourselves. */
+    if (Tracer_trace(self, frame, what, arg_unused) == RET_OK) {
+        return PyObject_GetAttrString((PyObject*)self, "pytrace");
+    }
+
+done:
+    return NULL;
 }
 
 static PyObject *
-Tracer_start(Tracer *self, PyObject *args)
+Tracer_start(Tracer *self, PyObject *args_unused)
 {
     PyEval_SetTrace((Py_tracefunc)Tracer_trace, (PyObject*)self);
     self->started = 1;
     self->tracing_arcs = self->arcs && PyObject_IsTrue(self->arcs);
     self->last_line = -1;
 
-    return Py_BuildValue("");
+    /* start() returns a trace function usable with sys.settrace() */
+    return PyObject_GetAttrString((PyObject*)self, "pytrace");
 }
 
 static PyObject *
-Tracer_stop(Tracer *self, PyObject *args)
+Tracer_stop(Tracer *self, PyObject *args_unused)
 {
     if (self->started) {
         PyEval_SetTrace(NULL, NULL);
@@ -505,6 +553,9 @@ Tracer_members[] = {
 
 static PyMethodDef
 Tracer_methods[] = {
+    { "pytrace",    (PyCFunction) Tracer_pytrace,       METH_VARARGS,
+            PyDoc_STR("A trace function compatible with sys.settrace()") },
+
     { "start",      (PyCFunction) Tracer_start,         METH_VARARGS,
             PyDoc_STR("Start the tracer") },
 
